wit_bindgen_rust::import!({paths: ["../../wit/http.wit"]});
use self::http::{
    send_http_request, HttpRequest as HttpRequestWit, HttpResponse as HttpResponseWit,
    HttpStatus as HttpStatusWit,
};
pub use self::http::{HttpMethod as HttpMethodWit, HttpRequestError};
use anyhow::Result;
use base64::{engine::general_purpose, Engine as _};
use serde::{Deserialize, Serialize, Serializer};
use std::collections::HashMap;
use std::convert::TryFrom;
use std::io::Stdin;

///////////////////////////////////////////////
/// HTTP Request
///////////////////////////////////////////////

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct HttpRequest {
    pub method: HttpMethod,
    pub uri: String,
    pub headers: HashMap<String, String>,
    pub params: Option<HashMap<String, String>>,
    pub body: Option<Vec<u8>>,
}

impl HttpRequest {
    pub fn new_from_stdin(reader: Stdin) -> Result<Self> {
        serde_json::from_reader::<Stdin, Self>(reader).map_err(anyhow::Error::new)
    }

    pub fn new_from_json(json: &str) -> Result<Self> {
        serde_json::from_str::<Self>(json).map_err(anyhow::Error::new)
    }

    pub fn send(&self) -> Result<HttpResponse, HttpRequestError> {
        let headers_slice: Vec<(&str, &str)> = self
            .headers
            .iter()
            .map(|(k, v)| (k.as_str(), v.as_str()))
            .collect::<Vec<(&str, &str)>>();

        let method: &HttpMethod = &self.method;
        let method = method.into();

        let http_request_wit = HttpRequestWit {
            body: self.body.as_deref(),
            headers: &headers_slice,
            method,
            params: &[],
            uri: &self.uri,
        };
        send_http_request(http_request_wit).map(|http_response_wit| http_response_wit.into())
    }

    pub fn body(&self) -> &Option<Vec<u8>> {
        &self.body
    }

    pub fn text(self) -> Option<String> {
        self.body.map(|b| String::from_utf8(b).unwrap())
    }
}

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(rename_all = "UPPERCASE")]
pub enum HttpMethod {
    Options,
    Get,
    Post,
    Put,
    Delete,
    Head,
    Patch,
}

impl From<&HttpMethod> for HttpMethodWit {
    fn from(http_method: &HttpMethod) -> Self {
        match http_method {
            HttpMethod::Options => HttpMethodWit::Options,
            HttpMethod::Get => HttpMethodWit::Get,
            HttpMethod::Post => HttpMethodWit::Post,
            HttpMethod::Put => HttpMethodWit::Put,
            HttpMethod::Delete => HttpMethodWit::Delete,
            HttpMethod::Head => HttpMethodWit::Head,
            HttpMethod::Patch => HttpMethodWit::Patch,
        }
    }
}

#[derive(Clone, Debug)]
pub struct HttpRequestBuilder {
    inner: HttpRequest,
}

impl HttpRequestBuilder {
    pub fn new(uri: String) -> Self {
        let inner = HttpRequest {
            uri,
            method: HttpMethod::Get,
            body: None,
            headers: HashMap::new(),
            params: None,
        };
        HttpRequestBuilder { inner }
    }

    pub fn uri(mut self, value: String) -> Self {
        self.inner.uri = value;
        self
    }

    pub fn method(mut self, value: HttpMethod) -> Self {
        self.inner.method = value;
        self
    }

    pub fn header(mut self, key: &str, value: &str) -> Self {
        self.inner.headers.insert(key.to_owned(), value.to_owned());
        self
    }

    pub fn param(mut self, key: String, value: String) -> Self {
        self.inner
            .params
            .get_or_insert_with(HashMap::new)
            .insert(key, value);
        self
    }

    pub fn body(mut self, value: Vec<u8>) -> Self {
        self.inner.body = Some(value);
        self
    }

    pub fn body_str(mut self, value: String) -> Self {
        self.inner.body = Some(value.into_bytes());
        self
    }

    pub fn build(self) -> HttpRequest {
        self.inner
    }
}

///////////////////////////////////////////////
/// HTTP Response
///////////////////////////////////////////////

#[derive(Serialize, Deserialize, Debug)]
pub struct HttpResponse {
    pub status: HttpStatus,
    pub headers: HashMap<String, String>,
    pub base64_body: bool,
    pub body: Option<Vec<u8>>,
}

impl HttpResponse {
    pub fn text(self) -> Option<String> {
        self.body.map(|content| String::from_utf8(content).unwrap())
    }
    pub fn status(&self) -> HttpStatus {
        self.status.clone()
    }
    pub fn body(&self) -> &Option<Vec<u8>> {
        &self.body
    }
    pub fn is_base64_body(&self) -> bool {
        self.base64_body
    }
    pub fn to_string(&self) -> Result<String> {
        serde_json::to_string(self).map_err(|err| err.into())
    }
}

impl From<HttpResponseWit> for HttpResponse {
    fn from(response_wit: HttpResponseWit) -> Self {
        let status = HttpStatus::try_from(response_wit.status).unwrap();
        let headers = response_wit
            .headers
            .iter()
            .map(|(k, v)| (k.clone(), v.clone()))
            .collect::<HashMap<_, _>>();
        let body = Some(response_wit.body.unwrap_or_default());

        HttpResponse {
            status,
            headers,
            body,
            base64_body: response_wit.base64_body,
        }
    }
}

impl From<&HttpResponse> for HttpResponseWit {
    fn from(response: &HttpResponse) -> Self {
        let status = response.status.clone().to_u16();
        let headers = response
            .headers
            .iter()
            .map(|(k, v)| (k.clone(), v.clone()))
            .collect::<Vec<(String, String)>>();
        HttpResponseWit {
            status,
            headers,
            body: response.body.clone(),
            base64_body: response.base64_body,
        }
    }
}

pub enum ResponseContent {
    Text(String),
    Binary(String),
}

impl From<Vec<u8>> for ResponseContent {
    fn from(s: Vec<u8>) -> ResponseContent {
        ResponseContent::Binary(general_purpose::STANDARD.encode(s))
    }
}

impl From<String> for ResponseContent {
    fn from(s: String) -> ResponseContent {
        ResponseContent::Text(s)
    }
}

pub struct HttpResponseBuilder {
    inner: HttpResponse,
}

impl HttpResponseBuilder {
    pub fn new(status: HttpStatus) -> Self {
        let inner = HttpResponse {
            status,
            headers: HashMap::new(),
            base64_body: false,
            body: None,
        };
        HttpResponseBuilder { inner }
    }

    pub fn ok() -> Self {
        Self::new(HttpStatus::Ok)
    }

    pub fn header(mut self, key: &str, value: &str) -> Self {
        self.inner.headers.insert(key.to_owned(), value.to_owned());
        self
    }

    pub fn body(mut self, value: ResponseContent) -> Self {
        let content = match value {
            ResponseContent::Text(text) => text.clone().into_bytes(),
            ResponseContent::Binary(text) => general_purpose::STANDARD.encode(text).into_bytes(),
        };
        self.inner.body = Some(content);
        self
    }

    pub fn build(self) -> HttpResponse {
        self.inner
    }
}

///////////////////////////////////////////////
/// HTTP Status
///////////////////////////////////////////////

#[repr(u16)]
#[derive(Deserialize, Clone, Debug)]
pub enum HttpStatus {
    // 1xx Informational Response
    Continue = 100,
    SwitchingProtocols = 101,
    Processing = 102,
    EarlyHints = 103,

    // 2xx Success
    Ok = 200,
    Created = 201,
    Accepted = 202,
    NonAuthoritativeInformation = 203,
    NoContent = 204,
    ResetContent = 205,
    PartialContent = 206,
    MultiStatus = 207,
    AlreadyReported = 208,
    IMUsed = 226,

    // 3xx Redirection
    MultipleChoices = 300,
    MovedPermanently = 301,
    Found = 302,
    SeeOther = 303,
    NotModified = 304,
    UseProxy = 305,
    Unused = 306,
    TemporaryRedirect = 307,
    PermanentRedirect = 308,

    // 4xx Client Error
    BadRequest = 400,
    Unauthorized = 401,
    PaymentRequired = 402,
    Forbidden = 403,
    NotFound = 404,
    MethodNotAllowed = 405,
    NotAcceptable = 406,
    ProxyAuthenticationRequired = 407,
    RequestTimeout = 408,
    Conflict = 409,
    Gone = 410,
    LengthRequired = 411,
    PreconditionFailed = 412,
    PayloadTooLarge = 413,
    URITooLong = 414,
    UnsupportedMediaType = 415,
    RangeNotSatisfiable = 416,
    ExpectationFailed = 417,
    ImATeapot = 418,
    MisdirectedRequest = 421,
    UnprocessableEntity = 422,
    Locked = 423,
    FailedDependency = 424,
    TooEarly = 425,
    UpgradeRequired = 426,
    PreconditionRequired = 428,
    TooManyRequests = 429,
    RequestHeaderFieldsTooLarge = 431,
    UnavailableForLegalReasons = 451,

    // 5xx Server Error
    InternalServerError = 500,
    NotImplemented = 501,
    BadGateway = 502,
    ServiceUnavailable = 503,
    GatewayTimeout = 504,
    HTTPVersionNotSupported = 505,
    VariantAlsoNegotiates = 506,
    InsufficientStorage = 507,
    LoopDetected = 508,
    NotExtended = 510,
    NetworkAuthenticationRequired = 511,
}

impl HttpStatus {
    /// Convert HttpStatus to u16 directly
    pub fn to_u16(self) -> u16 {
        self as u16
    }
}

impl Serialize for HttpStatus {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_u16(self.clone() as u16)
    }
}

impl TryFrom<HttpStatusWit> for HttpStatus {
    type Error = ();

    fn try_from(value: u16) -> Result<Self, Self::Error> {
        match value {
            100 => Ok(HttpStatus::Continue),
            101 => Ok(HttpStatus::SwitchingProtocols),
            102 => Ok(HttpStatus::Processing),

            200 => Ok(HttpStatus::Ok),
            201 => Ok(HttpStatus::Created),
            202 => Ok(HttpStatus::Accepted),
            203 => Ok(HttpStatus::NonAuthoritativeInformation),
            204 => Ok(HttpStatus::NoContent),
            205 => Ok(HttpStatus::ResetContent),
            206 => Ok(HttpStatus::PartialContent),

            300 => Ok(HttpStatus::MultipleChoices),
            301 => Ok(HttpStatus::MovedPermanently),
            302 => Ok(HttpStatus::Found),
            303 => Ok(HttpStatus::SeeOther),
            304 => Ok(HttpStatus::NotModified),
            305 => Ok(HttpStatus::UseProxy),
            307 => Ok(HttpStatus::TemporaryRedirect),
            308 => Ok(HttpStatus::PermanentRedirect),

            400 => Ok(HttpStatus::BadRequest),
            401 => Ok(HttpStatus::Unauthorized),
            402 => Ok(HttpStatus::PaymentRequired),
            403 => Ok(HttpStatus::Forbidden),
            404 => Ok(HttpStatus::NotFound),
            405 => Ok(HttpStatus::MethodNotAllowed),
            406 => Ok(HttpStatus::NotAcceptable),
            407 => Ok(HttpStatus::ProxyAuthenticationRequired),
            408 => Ok(HttpStatus::RequestTimeout),
            409 => Ok(HttpStatus::Conflict),
            410 => Ok(HttpStatus::Gone),
            411 => Ok(HttpStatus::LengthRequired),
            412 => Ok(HttpStatus::PreconditionFailed),
            413 => Ok(HttpStatus::PayloadTooLarge),
            414 => Ok(HttpStatus::URITooLong),
            415 => Ok(HttpStatus::UnsupportedMediaType),
            416 => Ok(HttpStatus::RangeNotSatisfiable),
            417 => Ok(HttpStatus::ExpectationFailed),
            418 => Ok(HttpStatus::ImATeapot),

            500 => Ok(HttpStatus::InternalServerError),
            501 => Ok(HttpStatus::NotImplemented),
            502 => Ok(HttpStatus::BadGateway),
            503 => Ok(HttpStatus::ServiceUnavailable),
            504 => Ok(HttpStatus::GatewayTimeout),
            505 => Ok(HttpStatus::HTTPVersionNotSupported),
            506 => Ok(HttpStatus::VariantAlsoNegotiates),
            507 => Ok(HttpStatus::InsufficientStorage),
            508 => Ok(HttpStatus::LoopDetected),
            510 => Ok(HttpStatus::NotExtended),
            511 => Ok(HttpStatus::NetworkAuthenticationRequired),

            _ => Err(()),
        }
    }
}
